/*

The MIT License

Copyright (c) 2008 Click to Flash Developers

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

*/


#import "Plugin.h"
#import "CTFWhitelistWindowController.h"

static NSString *sFlashOldMIMEType = @"application/x-shockwave-flash";
static NSString *sFlashNewMIMEType = @"application/futuresplash";
static NSString *sHostWhitelistDefaultsKey = @"ClickToFlash.whitelist";

static const CGFloat sClickToPlayAreaFraction = 0.8;
static NSString *sFakeFlashLogoFontName = @"Times-Roman";
static const CGFloat sFakeFlashLogoFontSize = 0.6;
static const CGFloat sFakeFlashLogoPositionX = 0.15;
static const CGFloat sFakeFlashLogoPositionY = 0.31;

@interface CTFClickToFlashPlugin (Internal)
- (void) _convertTypesForContainer;
- (void) _drawBackground;
- (void) _drawClickToPlayImage;
- (BOOL) _isOptionPressed;
- (BOOL) _isHostWhitelisted;
- (NSMutableArray *)_hostWhitelist;
- (void) _addHostToWhitelist;
- (void) _removeHostFromWhitelist;
@end


@implementation CTFClickToFlashPlugin


#pragma mark -
#pragma mark Class Methods

+ (NSView *)plugInViewWithArguments:(NSDictionary *)arguments
{
    return [[[self alloc] initWithArguments:arguments] autorelease];
}


#pragma mark -
#pragma mark Initialization and Superclass Overrides

- (id) initWithArguments:(NSDictionary *)arguments
{
    self = [super init];
    if (self) {
        self.container = [arguments objectForKey:WebPlugInContainingElementKey];
    
        NSURL *base = [arguments objectForKey:WebPlugInBaseURLKey];
        if (base) {
            self.baseURL = base;
            if ([self _isHostWhitelisted] && ![self _isOptionPressed]) {
                [self performSelector:@selector(_convertTypesForContainer) withObject:nil afterDelay:0];
            }
        }

        if (![NSBundle loadNibNamed:@"ContextualMenu" owner:self])
            NSLog(@"Could not load conextual menu plugin");

        NSDictionary *attributes = [arguments objectForKey:WebPlugInAttributesKey];
        if (attributes != nil) {
            NSString *src = [attributes objectForKey:@"src"];
            if (src)
                [self setToolTip:src];
            else {
                src = [attributes objectForKey:@"data"];
                if (src)
                    [self setToolTip:src];
            }
        }
    }

    return self;
}


- (void) dealloc
{
    self.container = nil;
    self.baseURL = nil;

    if (trackingArea) {
        [self removeTrackingArea:trackingArea];
        [trackingArea release];
    }

    [_whitelistWindowController release];
    [super dealloc];
}


- (void) drawRect:(NSRect)rect
{
    [self _drawBackground];
    [self _drawClickToPlayImage];
}


- (NSMenu *)menuForEvent:(NSEvent *)theEvent
{
    return [self menu];
}

- (BOOL) _isOptionPressed
{
    BOOL isOptionPressed = (([[NSApp currentEvent] modifierFlags] & NSAlternateKeyMask) != 0);
    return isOptionPressed;
}

- (void) mouseDown:(NSEvent *)event
{
    mouseIsDown = YES;
    mouseInside = YES;
    [self setNeedsDisplay:YES];

    // Track the mouse so that we can undo our pressed-in look if the user drags the mouse outside the view, and reinstate it if the user drags it back in.
    trackingArea = [[NSTrackingArea alloc] initWithRect:[self bounds]
                                                options:NSTrackingMouseEnteredAndExited | NSTrackingActiveInKeyWindow | NSTrackingEnabledDuringMouseDrag
                                                  owner:self
                                               userInfo:nil];
    [self addTrackingArea:trackingArea];
}

- (void)mouseEntered:(NSEvent *)event
{
    mouseInside = YES;
    [self setNeedsDisplay:YES];
}
- (void)mouseExited:(NSEvent *)event
{
    mouseInside = NO;
    [self setNeedsDisplay:YES];
}

- (void) mouseUp:(NSEvent *)event
{
    mouseIsDown = NO;
    // Display immediately because we don't want to end up drawing after we've swapped in the Flash movie.
    [self display];

    // We're done tracking.
    [self removeTrackingArea:trackingArea];
    [trackingArea release];
    trackingArea = nil;

    if (mouseInside) {
        if (self.baseURL && (([[NSApp currentEvent] modifierFlags] & NSAlternateKeyMask) == NSAlternateKeyMask) && ![self _isHostWhitelisted]) {
            NSString *title = [NSString stringWithFormat:NSLocalizedString(@"Add %@ to the white list?", @"Add %@ to the white list?"), [self.baseURL host]];
            int val = NSRunAlertPanel(title,
                                      NSLocalizedString(@"Always load flash for this site?", @"Always load flash for this site?"),
                                      @"Always Load", @"Don't Load", nil);
            if (!val) {
                return;
            }
            
            NSMutableArray *hostWhitelist = [[[[NSUserDefaults standardUserDefaults] stringArrayForKey:sHostWhitelistDefaultsKey] mutableCopy] autorelease];
            if (hostWhitelist) {
                [hostWhitelist addObject:[self.baseURL host]];
            } else {
                hostWhitelist = [NSMutableArray arrayWithObject:[self.baseURL host]];
            }
            [[NSUserDefaults standardUserDefaults] setObject:hostWhitelist forKey:sHostWhitelistDefaultsKey];
        }
        
        [self _convertTypesForContainer];
    }
}

- (BOOL) _isHostWhitelisted
{
    NSArray *hostWhitelist = [[NSUserDefaults standardUserDefaults] stringArrayForKey:sHostWhitelistDefaultsKey];
    return hostWhitelist && [hostWhitelist containsObject:[self.baseURL host]];
}

- (NSMutableArray *)_hostWhitelist
{
    NSMutableArray *hostWhitelist = [[[[NSUserDefaults standardUserDefaults] stringArrayForKey:sHostWhitelistDefaultsKey] mutableCopy] autorelease];
    if (hostWhitelist == nil) {
        hostWhitelist = [NSMutableArray array];
    }
    return hostWhitelist;
}

- (void) _addHostToWhitelist
{
    NSMutableArray *hostWhitelist = [self _hostWhitelist];
    [hostWhitelist addObject:[self.baseURL host]];
    [[NSUserDefaults standardUserDefaults] setObject:hostWhitelist forKey:sHostWhitelistDefaultsKey];
}

- (void) _removeHostFromWhitelist
{
    NSMutableArray *hostWhitelist = [self _hostWhitelist];
    [hostWhitelist removeObject:[self.baseURL host]];
    [[NSUserDefaults standardUserDefaults] setObject:hostWhitelist forKey:sHostWhitelistDefaultsKey];
}

#pragma mark -
#pragma mark Contextual menu

- (BOOL)validateMenuItem:(NSMenuItem *)menuItem
{
    BOOL enabled = YES;
    SEL action = [menuItem action];
    if (action == @selector(addToWhitelist:))
    {
        if ([self _isHostWhitelisted])
            enabled = NO;
    }
    else if (action == @selector(removeFromWhitelist:))
    {
        if (![self _isHostWhitelisted])
            enabled = NO;
    }
    
    return enabled;
}

- (IBAction)addToWhitelist:(id)sender
{
    if ([self _isHostWhitelisted])
        return;
    
    if ([self _isOptionPressed])
    {
        [self _addHostToWhitelist];
        [self _convertTypesForContainer];
        return;
    }
    
    NSString *title = NSLocalizedString(@"Always load flash for this site?", @"Always load flash for this site?");
    NSString *message = [NSString stringWithFormat:NSLocalizedString(@"Add %@ to the white list?", @"Add %@ to the white list?"), [self.baseURL host]];

    NSAlert *alert = [[[NSAlert alloc] init] autorelease];
    [alert addButtonWithTitle:NSLocalizedString(@"Add to white list", @"Add to white list")];
    [alert addButtonWithTitle:NSLocalizedString(@"Cancel", @"Cancel")];
    [alert setMessageText:title];
    [alert setInformativeText:message];
    [alert setAlertStyle:NSInformationalAlertStyle];
    [alert beginSheetModalForWindow:[self window]
                      modalDelegate:self
                     didEndSelector:@selector(addToWhitelistAlertDidEnd:returnCode:contextInfo:)
                        contextInfo:nil];
}

- (void)addToWhitelistAlertDidEnd:(NSAlert *)alert returnCode:(int)returnCode contextInfo:(void *)contextInfo
{
    if (returnCode == NSAlertFirstButtonReturn)
    {
        [self _addHostToWhitelist];
        [self _convertTypesForContainer];
    }
}

- (IBAction)removeFromWhitelist:(id)sender
{
    if (![self _isHostWhitelisted])
        return;
    
    NSString *title = NSLocalizedString(@"Remove from white list?", @"Remove from white list?");
    NSString *message = [NSString stringWithFormat:NSLocalizedString(@"Remove %@ from the white list?", @"Remove %@ from the white list?"), [self.baseURL host]];
    
    NSAlert *alert = [[[NSAlert alloc] init] autorelease];
    [alert addButtonWithTitle:NSLocalizedString(@"Remove from white list", @"Remove from white list")];
    [alert addButtonWithTitle:NSLocalizedString(@"Cancel", @"Cancel")];
    [alert setMessageText:title];
    [alert setInformativeText:message];
    [alert setAlertStyle:NSInformationalAlertStyle];
    [alert beginSheetModalForWindow:[self window]
                      modalDelegate:self
                     didEndSelector:@selector(removeFromWhitelistAlertDidEnd:returnCode:contextInfo:)
                        contextInfo:nil];
}

- (void)removeFromWhitelistAlertDidEnd:(NSAlert *)alert returnCode:(int)returnCode contextInfo:(void *)contextInfo
{
    if (returnCode == NSAlertFirstButtonReturn)
    {
        [self _removeHostFromWhitelist];
    }
}

- (IBAction)editWhitelist:(id)sender
{
    if (_whitelistWindowController == nil)
    {
        _whitelistWindowController = [[CTFWhitelistWindowController alloc] init];
    }
    [_whitelistWindowController showWindow:self];
}

- (IBAction)copyFlashMovieURLToGeneralPasteboard:(id)sender
{
    NSString *movieURLString = nil;
    NSURL *movieURL = nil;

    DOMElement *element = self.container;
    if ([element.tagName isEqualToString:@"EMBED"])
        movieURLString = [element getAttribute:@"src"];
    else if ([element.tagName isEqualToString:@"OBJECT"])
        movieURLString = [element getAttribute:@"data"];

    if (movieURLString)
        movieURL = [NSURL URLWithString:movieURLString relativeToURL:self.baseURL];

    if (movieURL) {
        NSPasteboard *pboard = [NSPasteboard generalPasteboard];
        [pboard declareTypes:[NSArray arrayWithObject:NSURLPboardType] owner:self];
        [movieURL writeToPasteboard:pboard];
    }
}

- (IBAction)loadFlash:(id)sender
{
    [self _convertTypesForContainer];
}

#pragma mark -
#pragma mark Drawing

- (void) _drawBackground
{
    NSRect selfBounds  = [self bounds];

    NSRect fillRect   = NSInsetRect(selfBounds, 1.0, 1.0);
    NSRect strokeRect = selfBounds;

    NSColor *startingColor = [NSColor colorWithDeviceWhite:1.0 alpha:0.15];
    NSColor *endingColor = [NSColor colorWithDeviceWhite:0.0 alpha:0.15];
    NSGradient *gradient = [[NSGradient alloc] initWithStartingColor:startingColor endingColor:endingColor];
    
    // Draw fill
    // When the mouse is up or outside the view, we want a convex look, so we draw the gradient downward (90+180=270 degrees).
    // When the mouse is down and inside the view, we want a concave look, so we draw the gradient upward (90 degrees).
    [gradient drawInBezierPath:[NSBezierPath bezierPathWithRect:fillRect] angle:90.0 + ((mouseIsDown && mouseInside) ? 0.0 : 180.0)];

    // Draw stroke
    [[NSColor colorWithCalibratedWhite:0.0 alpha:0.50] set];
    [NSBezierPath setDefaultLineWidth:2.0];
    [NSBezierPath setDefaultLineCapStyle:NSSquareLineCapStyle];
    [[NSBezierPath bezierPathWithRect:strokeRect] stroke];

    [gradient release];
}

- (void) _drawClickToPlayImage
{
    NSRect selfBounds  = [self bounds];

    NSGraphicsContext *context = [NSGraphicsContext currentContext];
    [context saveGraphicsState];

    // We will draw the click-to-play symbol centered inside an imaginary centered square. The square fits all the way across at least one dimension of the view (i.e., we proportionally scale it up to fit the view).
    // We scale so that 1.0 is the square's size. We translate to set our origin at the square's lower-left corner.
    NSAffineTransform *transform = [NSAffineTransform transform];
    CGFloat squareSize = MIN(selfBounds.size.width, selfBounds.size.height);
    [transform translateXBy:(selfBounds.size.width  - squareSize) / 2.0
                        yBy:(selfBounds.size.height - squareSize) / 2.0];
    [transform scaleBy:squareSize];
    // We don't want to draw the symbol at the full size of the square, so we shrink it a bit (keeping it centered).
    [transform translateXBy:(1.0 - sClickToPlayAreaFraction) / 2.0
                        yBy:(1.0 - sClickToPlayAreaFraction) / 2.0];
    [transform scaleBy:sClickToPlayAreaFraction];
    [transform concat];

    // If you need help visualizing the transformation, add code here to draw a unit square.

    NSBezierPath *playSymbol = [NSBezierPath bezierPath];
    NSBezierPath *glyphsPath = nil;

    // This is a simple right-pointing triangle that occupies the entire unit square.
    [playSymbol moveToPoint:NSZeroPoint];
    [playSymbol lineToPoint:(NSPoint){ 1.0, 0.5 }];
    [playSymbol lineToPoint:(NSPoint){ 0.0, 1.0 }];
    [playSymbol closePath];

    // Now to produce a florin (option-f) character.
    // We'll cut this out from the middle of the triangle later on.
    NSString *florinString = @"\u0192";

    NSFont *font = [NSFont fontWithName:sFakeFlashLogoFontName size:sFakeFlashLogoFontSize];
    NSDictionary *attributes = [NSDictionary dictionaryWithObjectsAndKeys:
        font, NSFontAttributeName,
        nil];
    NSTextStorage *textStorage = [[[NSTextStorage alloc] initWithString:florinString attributes:attributes] autorelease];

    NSLayoutManager *layoutManager = [[NSLayoutManager alloc] init];
    [textStorage addLayoutManager:layoutManager];
    [layoutManager release];

    NSRange glyphRange = [layoutManager glyphRangeForCharacterRange:(NSRange){ 0UL, [florinString length] } actualCharacterRange:NULL];
    NSGlyph *florinGlyphs = malloc(sizeof(NSGlyph) * glyphRange.length);
    if (florinGlyphs != NULL) {
        [layoutManager getGlyphs:florinGlyphs range:glyphRange];

        glyphsPath = [NSBezierPath bezierPath];
        [glyphsPath moveToPoint:(NSPoint){ sFakeFlashLogoPositionX, sFakeFlashLogoPositionY }];
        [glyphsPath appendBezierPathWithGlyphs:florinGlyphs count:glyphRange.length inFont:font];

        free(florinGlyphs);
    }

    // We have both pieces. Now to assemble them.
    NSBezierPath *bezierPath = [NSBezierPath bezierPath];
    [bezierPath appendBezierPath:playSymbol];
    // To cut out the shape instead of simply drawing it on top of the play symbol, we must reverse the path.
    if (glyphsPath)
        [bezierPath appendBezierPath:[glyphsPath bezierPathByReversingPath]];

    NSColor *startingColor = [NSColor colorWithDeviceWhite:0.0 alpha:0.4];
    NSColor *endingColor = [NSColor colorWithDeviceWhite:0.0 alpha:0.6];

    NSGradient *gradient = [[[NSGradient alloc] initWithStartingColor:startingColor endingColor:endingColor] autorelease];
    [bezierPath setWindingRule:NSEvenOddWindingRule];
    // When the mouse is up or outside the view, we want a convex look, so we draw the gradient downward (90+180=270 degrees).
    // When the mouse is down and inside the view, we want a concave look, so we draw the gradient upward (90 degrees).
    [gradient drawInBezierPath:bezierPath angle:90.0 + ((mouseIsDown && mouseInside) ? 0.0 : 180.0)];

    [context restoreGraphicsState];
}


#pragma mark -
#pragma mark DOM Conversion

- (void) _convertTypesForElement:(DOMElement *)element
{
    NSString *type = [element getAttribute:@"type"];

    if ([type isEqualToString:sFlashOldMIMEType] || [type length] == 0) {
        [element setAttribute:@"type" value:sFlashNewMIMEType];
    }
}


- (void) _convertTypesForContainer
{
    DOMElement *newElement = (DOMElement *)[self.container cloneNode:YES];

    DOMNodeList *nodeList = nil;
    NSUInteger i;

    [self _convertTypesForElement:newElement];

    nodeList = [newElement getElementsByTagName:@"object"];
    for (i = 0; i < nodeList.length; i++) {
        [self _convertTypesForElement:(DOMElement *)[nodeList item:i]];
    }

    nodeList = [newElement getElementsByTagName:@"embed"];
    for (i = 0; i < nodeList.length; i++) {
        [self _convertTypesForElement:(DOMElement *)[nodeList item:i]];
    }

    // Just to be safe, since we are about to replace our containing element
    [[self retain] autorelease];
    
    [self.container.parentNode replaceChild:newElement oldChild:self.container];
    self.container = nil;
}


@synthesize container = _container;
@synthesize baseURL;

@end
